%--------------------------------------------------------------------
%LPV Robust Synthesis (Numerical Example Demonstration)
%Update:12/01/2014
%Shu Wang
%--------------------------------------------------------------------
clear all;
clc;
%% Model Construction

% G0 = tf(1,[75 1]);
% MM = [87.8 -86.4;108.2 -109.6];
% Grho = ss(G0*MM);

rho = [1:0.5:3];  % time varying parameter rho
Np=size(rho,2);

tau = 71 + 2.*rho;  % system parameter tau

K11 = 87 + 0.2.*rho.^2;   % system parameter K
K12 = -87.2 + 0.2.*rho.^2;   % system parameter K
K21 = 107.4 + 0.2.*rho.^2;   % system parameter K
K22 = -110.4+ 0.2.*rho.^2;   % system parameter K

Grho = ss;
for i=1:Np
    
    Ki = [K11(i) K12(i);K21(i) K22(i)];
    Agi = blkdiag(-1/tau(i),-1/tau(i));
    Bgi = 1/tau(i)*eye(2);
    Cgi = Ki;
    Dgi = zeros(2,2);

    Grho(:,:,i) = ss(Agi,Bgi,Cgi,Dgi);
end

AG = Grho.a;
BG = Grho.b;
CG = Grho.c;
DG = Grho.d;

dz1 = 0.05;
dz2 = 0.05;


%% Used to check the controller and the gain of the nominal model 
We = ss(tf([1 0.1],[2 10^(-5)])*eye(2));

Wu = ss(tf([1 10],[1 100])*eye(2)); % control input

systemnames = 'Grho Wu We';
inputvar = '[r1; r2; u1; u2]';
outputvar = '[ We; Wu; r1 - Grho(1);r2 - Grho(2)]';
input_to_We = '[r1 - Grho(1);r2 - Grho(2)]';
input_to_Wu = '[u1;u2]';
input_to_Grho = '[u1;u2]';
cleanupsysic = 'yes';
Picn = sysic;

%%

opt = lpvsynOptions;
opt.BackOffFactor = 1.2;
[Knom,GAMnom,INFOnom] = lpvsyneng(Picn,2,2);
% [Knom,GAMnom,INFOnom] = hinfsyn(Picn,2,2);

fprintf('induced gain with the nominal controller: %f \t \n', GAMnom)

AKnom = Knom.a;
BKnom = Knom.b;
CKnom = Knom.c;
DKnom = Knom.d;

%%
% systemnames = 'G Wu We';
% inputvar = '[r1; r2; u1; u2]';
% outputvar = '[ We; Wu; r1 - G(1);r2 - G(2)]';
% input_to_We = '[r1 - G(1);r2 - G(2)]';
% input_to_Wu = '[u1;u2]';
% input_to_G = '[u1;u2]';
% cleanupsysic = 'yes';
% Picnlti = sysic;
% 
% opt = lpvsynOptions;
% opt.BackOffFactor = 1.2;
% [Knomlti,GAMnomlti,INFOnomlti] = lpvsyneng(Picnlti,2,2);
% 
% Fl = lft(Picn,Knomlti);
% [GAMlpv,info] = lpvnorm(Fl);
%% connection with saturation

beta1 = 0.6;
beta2 = 0.3;
Beta = ss(blkdiag(beta1,beta2));


systemnames = 'Grho Wu We Beta';
inputvar = '[d1; d2; r1; r2; u1; u2]';
outputvar = '[ Beta(1); Beta(2); We; Wu; r1 - Grho(1);r2 - Grho(2)]';
input_to_We = '[r1 - Grho(1); r2 - Grho(2)]';
input_to_Wu = '[u1; u2]';
input_to_Beta = '[u1; u2]';
input_to_Grho = '[u1 - d1; u2 - d2]';
cleanupsysic = 'yes';
Pic = sysic;
%% specify input variables for demo (sat)

szDelta = [2 2];
szK = [2 2];

Hsat = ss(tf(1,[1 1]));
Pisat2 = [0, (1 + Hsat); (1 + Hsat'), -(2 + Hsat + Hsat')];
Psisat2 = jspec(Pisat2);

Psis{1} = ss(eye(4));

Psi2_11 = blkdiag(Psisat2(1,1),Psisat2(1,1));
Psi2_12 = blkdiag(Psisat2(1,2),Psisat2(1,2));
Psi2_21 = blkdiag(Psisat2(2,1),Psisat2(2,1));
Psi2_22 = blkdiag(Psisat2(2,2),Psisat2(2,2));
Psis{2} = [Psi2_11 Psi2_12;Psi2_21 Psi2_22];

Pisat3 = [0 1; 1 -2];
Psisat3 = jspec(Pisat3);

Psi3_11 = blkdiag(Psisat3(1,1),Psisat3(1,1));
Psi3_12 = blkdiag(Psisat3(1,2),Psisat3(1,2));
Psi3_21 = blkdiag(Psisat3(2,1),Psisat3(2,1));
Psi3_22 = blkdiag(Psisat3(2,2),Psisat3(2,2));
Psis{3} = [Psi3_11 Psi3_12;Psi3_21 Psi3_22];


%% specify input variables for demo (sat) ii

% szDelta = [2 2];
% szK = [2 2];
% 
% Psisat1 = ss(eye(2));
% 
% Hsat = ss(tf(1,[1 1]));
% Pisat2 = [0, (1 + Hsat); (1 + Hsat'), -(2 + Hsat + Hsat')];
% Psisat2 = jspec(Pisat2);
% 
% Pisat3 = [0 1; 1 -2];
% Psisat3 = jspec(Pisat3);
% 
% Psis{1} = blkdiag(Psisat1,Psisat1);
% 
% Psi2_11 = blkdiag(Psisat1(1,1),Psisat2(1,1));
% Psi2_12 = blkdiag(Psisat1(1,2),Psisat2(1,2));
% Psi2_21 = blkdiag(Psisat1(2,1),Psisat2(2,1));
% Psi2_22 = blkdiag(Psisat1(2,2),Psisat2(2,2));
% Psis{2} = [Psi2_11 Psi2_12;Psi2_21 Psi2_22];
% 
% Psi3_11 = blkdiag(Psisat1(1,1),Psisat3(1,1));
% Psi3_12 = blkdiag(Psisat1(1,2),Psisat3(1,2));
% Psi3_21 = blkdiag(Psisat1(2,1),Psisat3(2,1));
% Psi3_22 = blkdiag(Psisat1(2,2),Psisat3(2,2));
% Psis{3} = [Psi3_11 Psi3_12;Psi3_21 Psi3_22];
% 
% Psi4_11 = blkdiag(Psisat2(1,1),Psisat1(1,1));
% Psi4_12 = blkdiag(Psisat2(1,2),Psisat1(1,2));
% Psi4_21 = blkdiag(Psisat2(2,1),Psisat1(2,1));
% Psi4_22 = blkdiag(Psisat2(2,2),Psisat1(2,2));
% Psis{4} = [Psi4_11 Psi4_12;Psi4_21 Psi4_22];
% 
% Psi5_11 = blkdiag(Psisat3(1,1),Psisat1(1,1));
% Psi5_12 = blkdiag(Psisat3(1,2),Psisat1(1,2));
% Psi5_21 = blkdiag(Psisat3(2,1),Psisat1(2,1));
% Psi5_22 = blkdiag(Psisat3(2,2),Psisat1(2,2));
% Psis{5} = [Psi5_11 Psi5_12;Psi5_21 Psi5_22];


%% Initialization
% dimensions of G (replace G by Pic for demo)
szG = size(Pic);
if numel(szG) == 2
    szG = [szG 1];
end
nmod = szG(3); % # of models in the G array.

szU = sum(szDelta,1);
nw = szU(1);   % # of outputs of Delta (uncertaint inputs to G)
nv = szU(2);   % # of inputs to Delta

nu = szK(1);    % # of outputs of K (control inputs)
ny = szK(2);    % # of inputs of K (measurements)

nd = szG(2) - nw - nu ;   % # of disturbances
ne = szG(1) - nv - ny ;   % # of errors

% n IQCs
niqcs = numel(Psis);

% lambda for n IQCs
lam = zeros(niqcs,1);
lam(1) = 1;

M = blkdiag(eye(nv),-eye(nw));

% stopping tolerence
maxiter = 25;

gamold = inf;
gamtol = 0.05;

%% check the robust performance of the nominal controller before synthesis

systemnames = 'Grho Knom Wu We Beta';
inputvar = '[d1; d2; r1; r2]';
outputvar = '[Beta(1); Beta(2); We; Wu]';
input_to_We = '[r1 - Grho(1); r2 - Grho(2)]';
input_to_Wu = '[Knom(1); Knom(2)]';
input_to_Beta = '[Knom(1);Knom(2)]';
input_to_Grho = '[Knom(1) - d1; Knom(2) - d2]';
input_to_Knom = '[r1 - Grho(1); r2 - Grho(2)]';
cleanupsysic = 'yes';
Hic = sysic;

% options for bisection algorithm
GAMnomlb = GAMnom;  %lower bound
GAMnomub = 20*GAMnom;

tol = 1e-2; %tolerance for bisection

while GAMnomub - GAMnomlb > tol
    GAMnomtry = (GAMnomlb + GAMnomub)/2;

    %run analysis
    [tnom,info] = lpvrbperf(Hic,szDelta,Psis,GAMnomtry);
    %new bounds for bisection
    if tnom <= 0
        GAMnomub = GAMnomtry;
    else
        GAMnomlb = GAMnomtry;
    end
end

GAMnomopt = GAMnomub;
% run analysis using gamopt
[tnom,info] = lpvrbperf(Hic,szDelta,Psis,GAMnomopt);

fprintf('robust performance with the nominal controller: %f \t \n', GAMnomopt)
%% Iterations
for ii = 1:maxiter
%     ii = 2;
    fprintf('\n Iteration: %i \n', ii);
    
    %% create a single IQC from n IQCs
    if ii ~= 1
        Pi = ss(zeros(nv + nw,nv + nw));
        Scl = blkdiag(1/gamold*eye(nv),eye(nw));
        for jj = 1:niqcs
            Pi = Pi + lam(jj)*Scl*Psis{jj}'*M*Psis{jj}*Scl;
        end
    end
    %% facotorization of the single IQC
    if ii == 1
        Psi = Psis{1};
    else
        [Psi,M] = jspec(Pi); 
        invScl = blkdiag(gamold*eye(nv),eye(nw));
        Psi = invScl*Psi;
    end

    %% create Psi_plus by inverting w/w_tilde channels
    Apsi = Psi.a;
    Bpsi = Psi.b;
    Cpsi_vtil = Psi.c(1:nv,:);
    Cpsi_wtil = Psi.c(nv+1:nv+nw,:);
    Dpsi_vtil = Psi.d(1:nv,:);
    Dpsi_wtilv = Psi.d(nv+1:nv+nw,1:nv);
    Dpsi_wtilw = Psi.d(nv+1:nv+nw,nv+1:nv+nw);    

    nxpsi = size(Apsi,1);
    
    Psi_mtx = [Apsi, Bpsi; Cpsi_vtil, Dpsi_vtil; zeros(nw,nxpsi+nv), eye(nw)];

%     if ii == 1
%         T = [zeros(nv,nw), eye(nv); inv(Dpsi_wtilw), -Dpsi_wtilw\Dpsi_wtilv];
%     else
        T = [eye(nxpsi), zeros(nxpsi,nw+nv); zeros(nv,nxpsi+nw), eye(nv); 
            -Dpsi_wtilw\Cpsi_wtil, inv(Dpsi_wtilw), -Dpsi_wtilw\Dpsi_wtilv];
%     end

    
    Psi_plusmtx = Psi_mtx*T;
    
    Psi_plusa = Psi_plusmtx(1:nxpsi,1:nxpsi);
    Psi_plusb = Psi_plusmtx(1:nxpsi,nxpsi+1:nxpsi+nw+nv);
    Psi_plusc = Psi_plusmtx(nxpsi+1:nxpsi+nv+nw,1:nxpsi);
    Psi_plusd = Psi_plusmtx(nxpsi+1:nxpsi+nv+nw,nxpsi+1:nxpsi+nw+nv);
    
    Psi_plus = ss(Psi_plusa,Psi_plusb,Psi_plusc,Psi_plusd);
    
    
    %% create scaled system G_scl by lft of G and Psi_plus 
    
    G_scl = lft(Psi_plus,Pic,nv,nw); %(replace G by Pic for demo)
%     if ii ~=1
%     GK = lft(G_scl,KGrid);
%     norm(GK,inf)
%     end

    %% nominal LPV synthesis step 
    % This "backs off" the optimal gamma by BackOffFactor = 1.1000
    opt = lpvsynOptions;
    opt.BackOffFactor = 1.2;
    [Krob,GAMrob,INFOrob] = lpvsyneng(G_scl,ny,nu);
%     [KGrid,CLGrid,GAMGrid,INFOGrid] = hinfsyn(G_scl,ny,nu);
   
    fprintf('Gamma controller step: %f \t \n', GAMrob)
    
    %% robust performance analysis step
    %get analysis interconnection (replace G by Pic for demo)
    Nic = lft(Pic,Krob);
    
    % options for bisection algorithm
    gamlb = 0;  %lower bound
    
    if ii==1    %upper bound   
        gamub = GAMrob; 
    else
%         gamub = gamold; 
        gamub = GAMrob;
    end

    tol = 1e-2; %tolerance for bisection
%     tol = 0.05*gamub;
    
    while gamub-gamlb>tol
        gamtry = (gamlb+gamub)/2;
        
        %run analysis
        [t,info] = lpvrbperf(Nic,szDelta,Psis,gamtry);
        %new bounds for bisection
        if t <= 0
            gamub = gamtry;
        else
            gamlb = gamtry;
        end
    end
    
    gamopt = gamub;
    % run analysis using gamopt
    [t,info] = lpvrbperf(Nic,szDelta,Psis,gamopt);

    fprintf('Gamma analysis step: %f \t \n', gamopt)
    %% save results for next iteration
    lam = info.lam;   
    gamvec(ii) = gamopt;
    
    if gamold-gamopt < gamtol
        break;
    end
    gamold = gamopt;
    
end

AKrob = Krob.a;
BKrob = Krob.b;
CKrob = Krob.c;
DKrob = Krob.d;








